@@ -39,6 +39,7 @@ echo "Welcome to $TITLE"
 
 mount -o remount,rw / &>/dev/null 
 cleanup_runtime
+needed_pkgs_fs= # will hold extra needed packages for blockdevices/filesystems, populated when the Fs'es are created
 
 ### Set configuration values ###
 # note : you're free to use or ignore these in your procedure.  probably you want to use these variables to override defaults in your configure worker
@@ -97,6 +98,9 @@ fi
 
 load_module core
 [ "$module" != core -a "$module" != http ] && load_module "$module"
+# this is a workaround for bash <4.2, where associative arrays are inherently local,
+# so we must source these variables in the main scope
+source $LIB_CORE/libs/lib-blockdevices-filesystems.sh
 
 load_procedure "$module" "$procedure"
 
@@ -142,4 +146,4 @@ start_installer
 
 start_process
 
-stop_installer
+stop_installer $?
