@@ -36,6 +36,66 @@ TMP_FILESYSTEMS=$RUNTIME_DIR/aif-filesystems # Only used internally by this libr
 TMP_BLOCKDEVICES=$RUNTIME_DIR/aif-blockdata
 
 
+declare -A filesystem_programs=(["swap"]="mkswap" ["reiserfs"]="mkreiserfs" ["lvm-pv"]="pvcreate" ["lvm-vg"]="vgcreate" ["lvm-lv"]="lvcreate" ["dm_crypt"]="cryptsetup")
+for simple in ext2 ext3 ext4 nilfs2 xfs jfs vfat btrfs
+do
+	filesystem_programs+=([$simple]=mkfs.$simple)
+done
+
+declare -A label_programs=(["swap"]="swaplabel" ["xfs"]="xfs_admin" ["jfs"]="jfs_tune" ["reiserfs"]="reiserfstune" ["nilfs2"]="nilfs-tune" ["vfat"]="dosfslabel")
+for ext in ext2 ext3 ext4
+do
+	label_programs+=([$ext]=tune2fs)
+done
+
+# names for filesystems (which are shown to users in dialogs etc, don't use spaces!)
+declare -A filesystem_names=(["nilfs2"]="Nilfs2 (EXPERIMENTAL)" ["btrfs"]="Btrfs (EXPERIMENTAL)" ["vfat"]="vFat" ["dm_crypt"]="dm_crypt (LUKS)" ["lvm-pv"]="lvm Physical Volume" ["lvm-vg"]="lvm Volume Group" ["lvm-lv"]="lvm Logical Volume")
+for i in ext2 ext3 ext4 reiserfs xfs jfs swap
+do
+	name=$(echo $i | capitalize)
+	filesystem_names+=([$i]=$name)
+done
+
+# specify which filesystems can be stored on which blockdevices
+fs_on_raw=(swap ext2 ext3 ext4 reiserfs nilfs2 xfs jfs vfat dm_crypt lvm-pv btrfs)
+fs_on_lvm_lv=(swap ext2 ext3 ext4 reiserfs nilfs2 xfs jfs vfat dm_crypt btrfs)
+fs_on_dm_crypt=(swap ext2 ext3 ext4 reiserfs nilfs2 xfs jfs vfat lvm-pv btrfs)
+fs_on_lvm_pv=(lvm-vg)
+fs_on_lvm_vg=(lvm-lv)
+
+# the following is useful to do checks like check_is_in $fs_type ${fs_on[$part_type]} (or iterate it)
+# bash does not allow lists (array) in associative arrays, but this seems to work.
+# maybe it stores them as strings, which in this case is ok (no spaces in elements)
+declare -A fs_on
+fs_on[raw]=${fs_on_raw[@]}
+fs_on[lvm-lv]=${fs_on_lvm_lv[@]}
+fs_on[lvm-pv]=${fs_on_lvm_pv[@]}
+fs_on[lvm-vg]=${fs_on_lvm_vg[@]}
+fs_on[dm_crypt]=${fs_on_dm_crypt[@]}
+
+fs_mountable=(btrfs ext2 ext3 ext4 nilfs2 xfs jfs vfat reiserfs)
+fs_label_mandatory=('lvm-vg' 'lvm-lv' 'dm_crypt')
+fs_label_optional=('swap' 'ext2' 'ext3' 'ext4' 'reiserfs' 'nilfs2' 'xfs' 'jfs' 'vfat')
+
+# list needed packages per filesystem
+declare -A filesystem_pkg=(["lvm-pv"]="lvm2" ["xfs"]="xfsprogs" ["jfs"]="jfsutils" ["reiserfs"]="reiserfsprogs" ["nilfs2"]="nilfs-utils" ["vfat"]="dosfstools" ["dm_crypt"]="cryptsetup" ["btrfs"]=btrfs-progs-unstable)
+for i in ext2 ext3 ext4
+do
+	filesystem_pkg+=([$i]=e2fsprogs)
+done
+
+# returns which filesystems you can create based on the locally available utilities
+get_possible_fs () {
+	possible_fs=
+	for fs in "${!filesystem_programs[@]}"
+	do
+		which ${filesystem_programs[$fs]} &>/dev/null && possible_fs=("${possible_fs[@]}" $fs)
+	done
+	true
+}
+
+syslinux_supported_fs=('ext2' 'ext3' 'ext4' 'btrfs' 'vfat')
+supported_bootloaders=('syslinux' 'grub')
 
 # procedural code from quickinst functionized and fixed.
 # there were functions like this in the setup script too, with some subtle differences.  see below
@@ -124,145 +184,133 @@ $ANSWER_DEVICES"
 }
 
 
-# taken from setup script, modified for separator control
-# $1 set to 1 to echo a newline after device instead of a space (optional)
-# $2 extra things to echo for each device (optional)
-# $3 something to append directly after the device (optional)
-finddisks() {
-    workdir="$PWD"
-    cd /sys/block 
-    # ide devices 
-    for dev in $(ls | egrep '^hd'); do
-        if [ "$(cat $dev/device/media)" = "disk" ]; then
-            echo -n "/dev/$dev$3"
-            [ "$1" = 1 ] && echo || echo -n ' '
-            [ "$2" ] && echo $2
-        fi
-    done  
-    #scsi/sata devices, and virtio blockdevices (/dev/vd*)
-    for dev in $(ls | egrep '^[sv]d'); do
-        # TODO: what is the significance of 5? ASKDEV
-        if ! [ "$(cat $dev/device/type)" = "5" ]; then
-            echo -n "/dev/$dev$3"
-            [ "$1" = 1 ] && echo || echo -n ' '
-            [ "$2" ] && echo $2
-        fi
-    done  
-    # cciss controllers
-    if [ -d /dev/cciss ] ; then
-        cd /dev/cciss
-        for dev in $(ls | egrep -v 'p'); do
-            echo -n "/dev/cciss/$dev$3"
-            [ "$1" = 1 ] && echo || echo -n ' '
-            [ "$2" ] && echo $2
-        done
-    fi
-    # Smart 2 controllers
-    if [ -d /dev/ida ] ; then
-        cd /dev/ida
-        for dev in $(ls | egrep -v 'p'); do
-            echo -n"/dev/ida/$dev$3"
-            [ "$1" = 1 ] && echo || echo -n ' '
-            [ "$2" ] && echo $2
-        done
-    fi
-    cd "$workdir"
-}
-
-
-# getuuid(). taken and modified from setup. this can probably be more improved. return an exit code, rely on blkid's exit codes etc.
-# converts /dev/[hs]d?[0-9] devices to UUIDs
-#
 # parameters: device file
 # outputs:    UUID on success
 #             nothing on failure
-# returns:    nothing
-getuuid()
-{
-	[ -z "$1" ] && die_error "getuuid needs an argument"
-	[ "${1%%/[hs]d?[0-9]}" != "${1}" ] && echo "$(blkid -s UUID -o value ${1})"
+# returns:    0 if the attribute is found, 1 otherwise
+getuuid() {
+	[ -n "$1" -a -b "$1" ] || die_error "getuuid needs a device file argument"
+	uuid=$(blkid -s UUID -o value ${1})
+	echo $uuid
+	[ -n "$uuid" ]
 }
 
-
 # parameters: device file
 # outputs:    LABEL on success
 #             nothing on failure
-# returns:    nothing
-getlabel()
-{
-	[ -z "$1" ] && die_error "getlabel needs an argument"
-	[ "${1%%/[hs]d?[0-9]}" != "${1}" ] && echo "$(blkid -s LABEL -o value ${1})"
+# returns:   0 if the attribute is found, 1 otherwise
+getlabel() {
+	[ -n "$1" -a -b "$1" ] || die_error "getlabel needs a device file argument"
+	label=$(blkid -s LABEL -o value ${1})
+	echo "$label"
+	[ -n "$label" ]
+}
+
+# find partitionable blockdevices
+# $1 extra things to echo for each device (optional) (backslash escapes will get interpreted)
+finddisks() {
+	shopt -s nullglob
+
+	# Block Devices
+	for dev in /sys/block/*; do
+		if [[ -f $dev/device/type ]]; then
+			local type
+			read type < /sys/block/${dev##*/}/device/type
+			# Block Devices with size =< 0 may be an empty card reader
+			read size < /sys/block/${dev##*/}/size
+			# Type 5 is a ROM Device - Optical Drives
+			if [[ $type != 5 ]] && (( $size > 0 )); then
+				source "$dev/uevent"
+				echo -ne "/dev/$DEVNAME $1"
+				unset DEVNAME
+			fi
+		fi
+	done
+
+	# cciss controllers
+	for dev in /dev/cciss/*; do
+		if [[ $dev != *[[:digit:]]p[[:digit:]]* ]]; then
+			echo "$dev $1"
+		fi
+	done
+
+	# Smart 2 Controller
+	for dev in /dev/ida/*; do
+		if [[ $dev != *[[:digit:]]p[[:digit:]]* ]]; then
+			echo "$dev $1"
+		fi
+	done
+
+	shopt -u nullglob
 }
 
 
-# taken from setup script, slightly optimized and modified for separator control
-# $1 set to 1 to echo a newline after partition instead of a space (optional)
-# $2 extra things to echo for each partition (optional)
-# $3 something to append directly after the partition (optional) TODO: refactor code so there's a space in between, merge $2 and $3. use echo -e to print whatever user wants
-findblockdevices() {
-	workdir="$PWD"
-	for devpath in $(finddisks)
-	do
-		disk=$(echo $devpath | sed 's|.*/||')
-		echo -n "/dev/$disk$3"
-		[ "$1" = 1 ] && echo || echo -n ' '
-		[ "$2" ] && echo $2
-		cd /sys/block/$disk   
-		for part in $disk*
-		do
-			# check if not already assembled to a raid device.  TODO: what is the significance of the 5? ASKDEV
-			if ! [ "$(grep $part /proc/mdstat 2>/dev/null)" -o "$(fstype 2>/dev/null </dev/$part | grep lvm2)" -o "$(sfdisk -c /dev/$disk $(echo $part | sed -e "s#$disk##g") 2>/dev/null | grep "5")" ]
-			then
-				if [ -d $part ]
-				then  
-					echo -n "/dev/$part$3"
-					[ "$1" = 1 ] && echo || echo -n ' '
-					[ "$2" ] && echo $2
-				fi
+# find usable blockdevices: RAID + LVM volumes, partitioned and unpartitioned devices
+# Exclude devices/partitions that are part of a RAID or LVM volume
+# Exclude root block devices (ex. sda) that are partitioned
+# $1 extra things to echo for each device (optional) (backslash escapes will get interpreted)
+find_usable_blockdevices() {
+	shopt -s nullglob
+
+	local parts
+
+	# Cycle through all root block devices (sda, sdb, etc...)
+	# Look for partitions and include them only if they are not part of a
+	# RAID or LVM configuration. Also, exclude extended partitions
+	for devpath in $(finddisks); do
+		hidebldev=
+		unset parts
+
+		# Glob allows for following matches:
+		# /dev/sda -> /dev/sda1
+		# /dev/cciss/c0d1 -> /dev/cciss/c0d1p1
+		for dev in ${devpath}*[[:digit:]]*; do
+			local disk="${dev%%[[:digit:]]*}"
+			local partnum="${dev##*[[:alpha:]]}"
+
+			# Don't display extended partition (not logical parts)
+			[ "$(sfdisk -c "$disk" "$partnum" 2>/dev/null)" = 5 ] && continue;
+
+			# Don't list parts that are part of RAID or LVM volumes
+			# Although we strongly encourage users to setup LVM in AIF, we want to make
+			# life a bit easier for those who have setup LVM/softraid before running AIF
+			if grep -qsw "${dev##*/}" /proc/mdstat || { pvscan -s 2>/dev/null | grep -q "$dev"; }; then
+				hidebldev="True"
+			else
+				parts+=("$dev")
 			fi
 		done
-	done
-	# include any mapped devices
-	for devpath in $(ls /dev/mapper 2>/dev/null | grep -v control)
-	do
-		echo -n "/dev/mapper/$devpath$3"
-		[ "$1" = 1 ] && echo || echo -n ' '
-		[ "$2" ] && echo $2
-	done
-	# include any raid md devices
-	for devpath in $(ls -d /dev/md* | grep '[0-9]' 2>/dev/null)
-	do
-		if grep -qw $(echo $devpath /proc/mdstat | sed -e 's|/dev/||g')
-		then
-			echo -n "$devpath$3"
-			[ "$1" = 1 ] && echo || echo -n ' '
-			[ "$2" ] && echo $2
+
+		# If hidebldev is not set and we have no partitions then
+		# Echo root block device if root block device is not part of a LVM or RAID configuration
+		# Otherwise echo the partitions
+		if [[ -z $hidebldev ]] && (( ! ${#parts[@]} )); then
+			if ! grep -qsw "${devpath##*/}" /proc/mdstat && ! pvscan -s 2>/dev/null | grep -q "$devpath"; then
+				echo -ne "$devpath $1"
+			fi
+		elif (( ${#parts[@]} )); then
+			for part in ${parts[@]}; do
+				echo -ne "$part $1"
+			done
 		fi
 	done
-	# inlcude cciss controllers
-	if [ -d /dev/cciss ]
-	then
-		cd /dev/cciss
-		for dev in $(ls | egrep 'p')
-		do
-			echo -n "/dev/cciss/$dev$3"
-			[ "$1" = 1 ] && echo || echo -n ' '
-			[ "$2" ] && echo $2
-		done
-	fi
-	# inlcude Smart 2 controllers
-	if [ -d /dev/ida ]
-	then
-		cd /dev/ida
-		for dev in $(ls | egrep 'p')
-		do
-			echo -n "/dev/ida/$dev$3"
-			[ "$1" = 1 ] && echo || echo -n ' '
-			[ "$2" ] && echo $2
-		done
-	fi
 
-	cd "$workdir"
+	# mapped devices
+	for devpath in /dev/mapper/*; do
+		# Exclude /control directory and other non-block files (such as??)
+		if [[ -b $devpath ]]; then
+			echo -ne "$devpath $1"
+		fi
+	done
+
+	# raid md devices
+	for devpath in /dev/md[0-9]*; do
+		if grep -qw "${devpath//\/dev\//}" /proc/mdstat; then
+			echo -ne "$devpath $1"
+		fi
+	done
+
+	shopt -u nullglob
 }
 
 
@@ -279,58 +327,53 @@ EOF
 # TODO: $1 is what?? ASKDEV
 # taken from setup. slightly edited.
 mapdev() {
-    partition_flag=0
-    device_found=0
-    devs=$( grep -v fd $TMP_DEV_MAP | sed 's/ *\t/ /' | sed ':a;$!N;$!ba;s/\n/ /g')
-    linuxdevice=$(echo $1 | cut -b1-8)
-    if [ "$(echo $1 | egrep '[0-9]$')" ]; then
-        # /dev/hdXY
-        pnum=$(echo $1 | cut -b9-)
-        pnum=$(($pnum-1))
-        partition_flag=1
-    fi
-    for  dev in $devs
-    do
-        if [ "(" = $(echo $dev | cut -b1) ]; then
-        grubdevice="$dev"
-        else
-        if [ "$dev" = "$linuxdevice" ]; then
-            device_found=1
-            break
-        fi
-       fi
-    done
-    if [ "$device_found" = "1" ]; then
-        if [ "$partition_flag" = "0" ]; then
-            echo "$grubdevice"
-        else
-            grubdevice_stringlen=${#grubdevice}
-            grubdevice_stringlen=$(($grubdevice_stringlen - 1))
-            grubdevice=$(echo $grubdevice | cut -b1-$grubdevice_stringlen)
-            echo "$grubdevice,$pnum)"
-        fi
-    else
-        echo "DEVICE NOT FOUND" >&2
-        return 2
-    fi
+	partition_flag=0
+	device_found=0
+	devs=$( grep -v fd $TMP_DEV_MAP | sed 's/ *\t/ /' | sed ':a;$!N;$!ba;s/\n/ /g')
+	linuxdevice=$(echo $1 | cut -b1-8)
+	if [ "$(echo $1 | egrep '[0-9]$')" ]; then
+		# /dev/hdXY
+		pnum=$(echo $1 | cut -b9-)
+		pnum=$(($pnum-1))
+		partition_flag=1
+	fi
+	for dev in $devs
+	do
+		if [ "(" = $(echo $dev | cut -b1) ]; then
+			grubdevice="$dev"
+		else
+			if [ "$dev" = "$linuxdevice" ]; then
+				device_found=1
+				break
+			fi
+		fi
+	done
+	if [ "$device_found" = "1" ]; then
+		if [ "$partition_flag" = "0" ]; then
+			echo "$grubdevice"
+		else
+			grubdevice_stringlen=${#grubdevice}
+			grubdevice_stringlen=$(($grubdevice_stringlen - 1))
+			grubdevice=$(echo $grubdevice | cut -b1-$grubdevice_stringlen)
+			echo "$grubdevice,$pnum)"
+		fi
+		return 0
+	else
+		return 1
+	fi
 }
 
 
 
-# auto_fstab(). taken from setup
 # preprocess fstab file
 # comments out old fields and inserts new ones
-# according to partitioning/formatting stage
-#
+# according to $TMP_FSTAB, set through process_filesystem calls
 target_configure_fstab()
 {
-	if [ -f $TMP_FSTAB ]
-	then
-		# comment out stray /dev entries
-		sed -i 's/^\/dev/#\/dev/g' $var_TARGET_DIR/etc/fstab
-		# append entries from new configuration
-		sort $TMP_FSTAB >>$var_TARGET_DIR/etc/fstab
-	fi
+	[ -f $TMP_FSTAB ] || return 0 # we can't do anything, but not really a failure
+	sed -i 's!^\(/dev/\|LABEL=\|UUID=\)!#\1!' $var_TARGET_DIR/etc/fstab || return 1
+	sort $TMP_FSTAB >> $var_TARGET_DIR/etc/fstab || return 1
+	return 0
 }
 
 
@@ -358,7 +401,7 @@ partition()
 	# format: each line=1 part.  <start> <size> <id> <bootable>[ <c,h,s> <c,h,s>]
 
 	read -r -a fsspecs <<< "$STRING"  # split up like this otherwise '*' will be globbed. which usually means an entry containing * is lost
-
+	sfdisk_input=
 	for fsspec in "${fsspecs[@]}"; do
 		fssize=$(echo $fsspec | tr -d ' ' | cut -f1 -d:)
 		fssize_spec=",$fssize"
@@ -383,14 +426,14 @@ partition()
 	sfdisk -D $DEVICE -uM >$LOG 2>&1 <<EOF
 $sfdisk_input
 EOF
-    if [ $? -gt 0 ]; then
-        notify "Error partitioning $DEVICE (see $LOG for details)"
-        printk on
-        return 1
-    fi
-    printk on
+	if [ $? -gt 0 ]; then
+		notify "Error partitioning $DEVICE (see $LOG for details)"
+		printk on
+		return 1
+	fi
+	printk on
 
-    return 0
+	return 0
 }
 
 
@@ -410,27 +453,54 @@ process_disk ()
 	partition $1 "$2"
 }
 
-# $1 fs_string
-# $2 decode yes/no
+# fills variables with info found in $TMP_BLOCKDEVICES
+# $1 partition to look for
+# $2 value to replace "no_foo" values with (optional) (can be '')
+getpartinfo () {
+	part=$1
+	declare part_escaped=${part//\//\\/} # escape all slashes otherwise awk complains
+	declare part_escaped=${part_escaped/+/\\+} # escape the + sign too
+	part_type=$( awk "/^$part_escaped / {print \$2}" $TMP_BLOCKDEVICES)
+	part_label=$(awk "/^$part_escaped / {print \$3}" $TMP_BLOCKDEVICES)
+	fs=$(        awk "/^$part_escaped / {print \$4}" $TMP_BLOCKDEVICES)
+	if [ -n "${2+2}" ]; then # checking if a var is defined, in bash.
+		[ "$part_label" = no_label ] && part_label=$2
+		[ "$fs"         = no_fs    ] && fs=$2
+	fi
+}
+
+# $1 fs_string (either a ; delimited string, or 'no_fs')
+# $2 value to replace empty values with (optional or '' for no effect)
+# $3 value to replace "no_foo" values with (optional) (can be '')
+#    when given, __ will be translated to ' ' as well
 parse_filesystem_string ()
 {
 	fs="$1"
-	fs_type=`       cut -d ';' -f 1 <<< $fs`
-	fs_create=`     cut -d ';' -f 2 <<< $fs`
-	fs_mountpoint=` cut -d ';' -f 3 <<< $fs`
-	fs_mount=`      cut -d ';' -f 4 <<< $fs`
-	fs_opts=`       cut -d ';' -f 5 <<< $fs`
-	fs_label=`      cut -d ';' -f 6 <<< $fs`
-	fs_params=`     cut -d ';' -f 7 <<< $fs`
-	if [ "$2" == 'yes' ]; then
+	fs_type=`       cut -s -d ';' -f 1 <<< $fs`
+	fs_create=`     cut -s -d ';' -f 2 <<< $fs`
+	fs_mountpoint=` cut -s -d ';' -f 3 <<< $fs`
+	fs_mount=`      cut -s -d ';' -f 4 <<< $fs`
+	fs_opts=`       cut -s -d ';' -f 5 <<< $fs`
+	fs_label=`      cut -s -d ';' -f 6 <<< $fs`
+	fs_params=`     cut -s -d ';' -f 7 <<< $fs`
+	if [ -n "$2" ]; then
+		[ -z "$fs_type"       ] && fs_type=$2
+		[ -z "$fs_create"     ] && fs_create=$2
+		[ -z "$fs_mountpoint" ] && fs_mountpoint=$2
+		[ -z "$fs_mount"      ] && fs_mount=$2
+		[ -z "$fs_opts"       ] && fs_opts=$2
+		[ -z "$fs_label"      ] && fs_label=$2
+		[ -z "$fs_params"     ] && fs_params=$2
+	fi
+	if [ -n "${3+3}" ]; then # checking if a var is defined, in bash.
 		fs_opts="${fs_opts//__/ }"
 		fs_params="${fs_params//__/ }"
-		[ "$fs_type"       = no_type       ] && fs_type=
-		[ "$fs_mountpoint" = no_mountpoint ] && fs_mountpoint=
-		[ "$fs_mount"      = no_mount      ] && fs_mount=
-		[ "$fs_opts"       = no_opts       ] && fs_opts=
-		[ "$fs_label"      = no_label      ] && fs_label=
-		[ "$fs_params"     = no_params     ] && fs_params=
+		[ "$fs_type"       = no_type       ] && fs_type=$3
+		[ "$fs_mountpoint" = no_mountpoint ] && fs_mountpoint=$3
+		[ "$fs_mount"      = no_mount      ] && fs_mount=$3
+		[ "$fs_opts"       = no_opts       ] && fs_opts=$3
+		[ "$fs_label"      = no_label      ] && fs_label=$3
+		[ "$fs_params"     = no_params     ] && fs_params=$3
 	fi
 }
 
@@ -478,6 +548,9 @@ process_filesystems ()
 				then
 					debug 'FS' "$fs_id ->Already done"
 				else
+					needs_pkg=${filesystem_pkg[$fs_type]}
+					[ -n "$needs_pkg" ] && check_is_in $needs_pkg "${needed_pkgs_fs[@]}" || needed_pkgs_fs+=($needs_pkg)
+
 					# We can't always do -b on the lvm VG. because the devicefile sometimes doesn't exist for a VG. vgdisplay to the rescue!
 					if [ "$part_type" = lvm-vg ] && vgdisplay $part | grep -q 'VG Name' # $part is a lvm VG and it exists. note that vgdisplay exists 0 when the requested vg doesn't exist.
 					then
@@ -503,7 +576,7 @@ process_filesystems ()
 		done < $TMP_FILESYSTEMS
 		[ $open_items -eq 0 ] && break
 	done
-	[ $open_items -eq 1 ] && show_warning "Filesystem/blockdevice processor problem" "Warning: Could not create all needed filesystems.  Either the underlying blockdevices didn't became available in 10 iterations, or process_filesystem failed" && returncode=1
+	[ $open_items -eq 1 ] && show_warning "Filesystem/blockdevice processor problem" "Warning: Could not create all needed filesystems.  Either the underlying blockdevices didn't appear in 10 iterations, or process_filesystem failed" && returncode=1
 
 
 
@@ -572,7 +645,7 @@ rollback_filesystems ()
 	# phase 2: destruct blockdevices listed in $BLOCK_DATA if they would exist already, in the correct order (first lvm LV, then VG, then PV etc)
 	# targets are device-mapper devices such as any lvm things, dm_crypt devices, etc and lvm PV's.
 
-	# Possible approach 1 (not implemented): for each target in $TMP_BLOCKDEVICES, check that it has no_fs or has a non-lvm/dm_crypt fs. (egrep -v ' lvm-pv;| lvm-vg;| lvm-lv;| dm_crypt;' ) and clean it 
+	# Possible approach 1 (not implemented): for each target in $TMP_BLOCKDEVICES, check that it has no_fs or has a non-lvm/dm_crypt fs. (egrep -v ' lvm-pv;| lvm-vg;| lvm-lv;| dm_crypt;' ) and clean it
 	#                      -> requires updating of underlying block device string when you clean something, on a copy of .block_data etc. too complicated
 	# Approach 2 : iterate over all targets in $TMP_BLOCKDEVICES as much as needed, until a certain limit, and in each loop check what can be cleared by looking at the real, live usage of / dependencies on the partition.
 	#                      -> easier (implemented)
@@ -687,6 +760,7 @@ rollback_filesystems ()
 	[ -n "$warnings" ] && inform "Rollback failed" disks 1 && show_warning "Rollback problems" "Some problems occurred while rolling back: $warnings.\n Thisk needs to be fixed before retrying disk/filesystem creation or restarting the installer" && return 1
 	inform "Rollback succeeded" disks 1
 	done_filesystems=
+	needed_pkgs_fs=
 	BLOCK_ROLLBACK_USELESS=1
 	return 0
 }
@@ -710,12 +784,12 @@ process_filesystem ()
 	debug 'FS' "process_filesystem $@"
 	local ret=0
 	part=$1
-	parse_filesystem_string "$2;$3;$4;$5;$6;$7;$8" yes
+	parse_filesystem_string "$2;$3;$4;$5;$6;$7;$8" '' ''
 
 	# Create the FS
 	if [ "$fs_create" = yes ]
 	then
-		if ! program=`get_filesystem_program $fs_type`
+		if ! check_is_in $fs_type "${!filesystem_programs[@]}"
 		then
 			show_warning "process_filesystem error" "Cannot determine filesystem program for $fs_type on $part.  Not creating this FS"
 			return 1
@@ -723,30 +797,28 @@ process_filesystem ()
 		[ -z "$fs_label" ] && [ "$fs_type" = lvm-vg -o "$fs_type" = lvm-pv ] && fs_label=default #TODO. implement the incrementing numbers label for lvm vg's and lv's
 
 		#TODO: health checks on $fs_params etc
-		case ${fs_type} in 
-			xfs)      mkfs.xfs -f $part           $fs_opts >$LOG 2>&1; ret=$? ;;
-			jfs)      yes | mkfs.jfs $part        $fs_opts >$LOG 2>&1; ret=$? ;;
-			reiserfs) yes | mkreiserfs $part      $fs_opts >$LOG 2>&1; ret=$? ;;
-			ext2)     mke2fs "$part"              $fs_opts >$LOG 2>&1; ret=$? ;;
-			ext3)     mke2fs -j $part             $fs_opts >$LOG 2>&1; ret=$? ;;
-			ext4)     mkfs.ext4 $part             $fs_opts >$LOG 2>&1; ret=$? ;; #TODO: installer.git uses mke2fs -t ext4 -O dir_index,extent,uninit_bg , which is best?
-			vfat)     mkfs.vfat $part             $fs_opts >$LOG 2>&1; ret=$? ;;
-			swap)     if [ -z "$fs_label" ]; then
-				  	mkswap $part          $fs_opts >$LOG 2>&1; ret=$?
-				  else
-				  	mkswap -L $fs_label $part $fs_opts >$LOG 2>&1; ret=$?
-				  fi
-				  ;;
-			dm_crypt) [ -z "$fs_params" ] && fs_params='-c aes-xts-plain -y -s 512';
-                      inform "Please enter your passphrase to encrypt the device (with confirmation)"
-			          cryptsetup $fs_params $fs_opts luksFormat -q $part >$LOG 2>&1 < /dev/tty ; ret=$? #hack to give cryptsetup the approriate stdin. keep in mind we're in a loop (see process_filesystems where something else is on stdin)
-                      inform "Please enter your passphrase to unlock the device"
-			          cryptsetup       luksOpen $part $fs_label >$LOG 2>&1 < /dev/tty; ret=$? || ( show_warning 'cryptsetup' "Error luksOpening $part on /dev/mapper/$fs_label" ) ;;
-			lvm-pv)   pvcreate $fs_opts $part              >$LOG 2>&1; ret=$? ;;
-			lvm-vg)   # $fs_params: list of PV's
-			          vgcreate $fs_opts $fs_label $fs_params      >$LOG 2>&1; ret=$? ;;
-			lvm-lv)   # $fs_params = size string (eg '5G')
-			          lvcreate -L $fs_params $fs_opts -n $fs_label `sed 's#/dev/mapper/##' <<< $part`   >$LOG 2>&1; ret=$? ;; #$fs_opts is usually something like -L 10G # Strip '/dev/mapper/' part because device file may not exist.  TODO: do i need to activate them?
+		program="${filesystem_programs[$fs_type]}"
+		case ${fs_type} in
+			xfs)
+				$program -f $part           $fs_opts >$LOG 2>&1; ret=$? ;;
+			jfs|reiserfs)
+				yes | $program $part        $fs_opts >$LOG 2>&1; ret=$? ;;
+			swap|ext2|ext3|ext4|nilfs2|vfat)
+				$program $part              $fs_opts >$LOG 2>&1; ret=$? ;;
+			dm_crypt)
+				[ -z "$fs_params" ] && fs_params='-c aes-xts-plain -y -s 512';
+				inform "Please enter your passphrase to encrypt the device (with confirmation)"
+				$program $fs_params $fs_opts luksFormat -q $part >$LOG 2>&1 < /dev/tty ; ret=$? #hack to give cryptsetup the approriate stdin. keep in mind we're in a loop (see process_filesystems where something else is on stdin)
+				inform "Please enter your passphrase to unlock the device"
+				$program luksOpen $part $fs_label >$LOG 2>&1 < /dev/tty; ret=$? || ( show_warning 'cryptsetup' "Error luksOpening $part on /dev/mapper/$fs_label" ) ;;
+			lvm-pv)
+				$program $fs_opts $part              >$LOG 2>&1; ret=$? ;;
+			lvm-vg)
+				# $fs_params: list of PV's
+				$program $fs_opts $fs_label $fs_params      >$LOG 2>&1; ret=$? ;;
+			lvm-lv)
+				# $fs_params = size string (eg '5G')
+				$program -L $fs_params $fs_opts -n $fs_label `sed 's#/dev/mapper/##' <<< $part`   >$LOG 2>&1; ret=$? ;; #$fs_opts is usually something like -L 10G # Strip '/dev/mapper/' part because device file may not exist.  TODO: do i need to activate them?
 			# don't handle anything else here, we will error later
 		esac
 		# The udevadm settle is a workaround for a bug/racecondition in cryptsetup. See:
@@ -758,24 +830,17 @@ process_filesystem ()
 		sleep 2
 	fi
 
-	if [ -n "$fs_label" ]
+	if [ -n "$fs_label" ] && check_is_in $fs_type "${!label_programs[@]}"
 	then
-		if [ "$fs_type" = xfs ]
-		then
-			xfs_admin -L $fs_label $part		>$LOG 2>&1; ret=$?
-		elif [ "$fs_type" = jfs ]
-		then
-			jfs_tune -L $fs_label $part		>$LOG 2>&1; ret=$?
-		elif [ "$fs_type" = reiserfs ]
-		then
-			reiserfstune -l $fs_label $part		>$LOG 2>&1; ret=$?
-		elif [ "$fs_type" = ext2 -o "$fs_type" = ext3 -o "$fs_type" = ext4 ]
-		then
-			tune2fs -L $fs_label $part		>$LOG 2>&1; ret=$?
-		elif [ "$fs_type" = vfat ]
-		then
-			dosfslabel $part $fs_label		>$LOG 2>&1; ret=$?
-		fi
+		program="${label_programs[$fs_type]}"
+		case ${fs_type} in
+			swap|xfs|jfs|nilfs2|ext2|ext3|ext4)
+				$program -L $fs_label $part		>$LOG 2>&1; ret=$?;;
+			reiserfs)
+				$program -l $fs_label $part		>$LOG 2>&1; ret=$?;;
+			vfat)
+				$program $part $fs_label		>$LOG 2>&1; ret=$?;;
+		esac
 		[ "$ret" -gt 0 ] && { show_warning "process_filesystem error" "Error setting label $fs_label on $part." ; return 1; }
 	fi
 
@@ -803,17 +868,9 @@ process_filesystem ()
 	then
 		case "$PART_ACCESS" in
 			label)
-				local _label="$(getlabel $part)"
-				if [ -n "${_label}" ]; then
-					part="LABEL=${_label}"
-				fi
-				;;
+				local label="$(getlabel $part)" && part="LABEL=$label";;
 			uuid)
-				local _uuid="$(getuuid $part)"
-				if [ -n "${_uuid}" ]; then
-					part="UUID=${_uuid}"
-				fi
-				;;
+				local uuid="$(getuuid $part)" && part="UUID=$uuid";;
 		esac
 		if ! grep -q "$part $fs_mountpoint $fs_type defaults 0 " $TMP_FSTAB 2>/dev/null #$TMP_FSTAB may not exist yet
 		then
@@ -833,26 +890,6 @@ process_filesystem ()
 }
 
 
-# $1 filesystem type
-get_filesystem_program ()
-{
-	[ -z "$1" ] && die_error "get_filesystem_program needs a filesystem id as \$1"
-	[ $1 = swap     ] && echo mkswap     && return 0
-	[ $1 = ext2     ] && echo mkfs.ext2  && return 0
-	[ $1 = ext3     ] && echo mkfs.ext3  && return 0
-	[ $1 = ext4     ] && echo mkfs.ext4  && return 0
-	[ $1 = reiserfs ] && echo mkreiserfs && return 0
-	[ $1 = xfs      ] && echo mkfs.xfs   && return 0
-	[ $1 = jfs      ] && echo mkfs.jfs   && return 0
-	[ $1 = vfat     ] && echo mkfs.vfat  && return 0
-	[ $1 = lvm-pv   ] && echo pvcreate   && return 0
-	[ $1 = lvm-vg   ] && echo vgcreate   && return 0
-	[ $1 = lvm-lv   ] && echo lvcreate   && return 0
-	[ $1 = dm_crypt ] && echo cryptsetup && return 0
-	return 1
-}
-
-
 # $1 blockdevice
 # $2 unit: B, KiB, kB, MiB, MB, GiB or GB.  defaults to B (we follow IEEE 1541-2002 )
 # output will be in $BLOCKDEVICE_SIZE
@@ -873,6 +910,7 @@ get_blockdevice_size ()
 	# BLOCKDEVICE_SIZE=$(($blocks/1024))
 
 	bytes=$((`fdisk -l $1 2>/dev/null | sed -n '2p' | cut -d' ' -f5`))
+	[[ $bytes = *[^0-9]* ]] && die_error "Could not parse fdisk -l output for $1"
 	[ $unit = B   ] && BLOCKDEVICE_SIZE=$bytes
 	[ $unit = KiB ] && BLOCKDEVICE_SIZE=$((bytes/2**10)) # /1024
 	[ $unit = kB  ] && BLOCKDEVICE_SIZE=$((bytes/10**3)) # /1000
@@ -880,23 +918,17 @@ get_blockdevice_size ()
 	[ $unit = MB  ] && BLOCKDEVICE_SIZE=$((bytes/10**6))
 	[ $unit = GiB ] && BLOCKDEVICE_SIZE=$((bytes/2**30))
 	[ $unit = GB  ] && BLOCKDEVICE_SIZE=$((bytes/10**9))
+	true
 }
 
 
 # $1 blockdevice (ex: /dev/md0 or /dev/sda1)
-# return true when blockdevice is an md raid, otherwise return a unset value
-mdraid_is-raid ()
-{
-    local israid
-    if [ -z $1 ]; then
-        # Don't call mdadm on empty blockdevice parameter!
-        israid=""
-    elif [ "$(mdadm --query $1 | cut -d':' -f2)" == " is not an md array" ]; then
-        israid=""
-    else
-        israid=true
-    fi
-    echo $israid
+# All MD RAID block devices have a major id of 9
+device_is_raid() {
+	[[ -b "$1" ]] || die_error "device_is_raid needs a blockdevice as \$1 ($1 given)"
+	[[ -f /proc/mdstast ]] || return 1
+	local devmajor=$(stat -c %t "$1")
+	(( devmajor == 9 ))
 }
 
 # $1 md raid blockdevice (ex: /dev/md0)
@@ -907,18 +939,22 @@ mdraid_is-raid ()
 # This procedure is used to determine the grub value for root, ex: (hd0,0)
 mdraid_slave0 ()
 {
-    echo "/dev/"$(ls -ldgGQ /sys/class/block/$(basename $1)/md/rd0 | cut -d'"' -f4 | cut -d'-' -f2)
+	echo "/dev/"$(ls -ldgGQ /sys/class/block/$(basename $1)/md/rd0 | cut -d'"' -f4 | cut -d'-' -f2)
 }
 
 # $1 md raid blockdevice (ex: /dev/md0)
 # return a list of array members from given md array
 # ex: /dev/md0 has slaves: "/dev/sda1 /dev/sdb2 /dev/sdc2"
-mdraid_all-slaves ()
+mdraid_all_slaves ()
 {
-    local slave=
-    local slaves=
-    for slave in $(ls /sys/class/block/$(basename $1)/slaves/); do
-        slaves=$slaves"/dev/"$slave" "
-    done
-    echo $slaves
+	shopt -s nullglob
+	local slave=
+	local slaves=
+	for slave in /sys/class/block/${1##*/}/slaves/*; do
+		source "$slave/uevent"
+		slaves="$slaves/dev/$DEVNAME "
+		unset DEVNAME
+	done
+	shopt -u nullglob
+	echo $slaves
 }
