@@ -26,7 +26,7 @@ load_module ()
 {
 	[ -z "$1" ] && die_error "load_module needs a module argument"
 	log "Loading module $1 ..."
-	path=$LIB_USER/"$1"
+	local path=$LIB_USER/"$1"
 	[ "$1" = core ] && path=$LIB_CORE
 	
 	for submodule in lib #procedure don't load procedures automatically!
@@ -37,13 +37,15 @@ load_module ()
 			[ "$1" = core ] && die_error "$path/${submodule}s does not exist. something is horribly wrong with this installation"
 		else
 			shopt -s nullglob
-			for i in "$path/${submodule}s"/*
+			local module
+			for module in "$path/${submodule}s"/*
 			do
 				# I have the habit of editing files while testing, don't source my backup files!
-				[[ "$i" == *~ ]] && continue
-
-				load_${submodule} "$1" "`basename "$i"`"
+				[[ "$module" == *~ ]] && continue
+				module=$(basename "$module")
+				load_${submodule} "$1" "$module"
 			done
+			shopt -u nullglob
 		fi
 	done
 			
@@ -59,11 +61,11 @@ load_procedure()
 	if [ "$1" = 'http' ]
 	then
 		log "Loading procedure $2 ..."
-		procedure=$RUNTIME_DIR/aif-procedure-downloaded-`basename $2`
+		local procedure=$RUNTIME_DIR/aif-procedure-downloaded-`basename $2`
 		wget "$2" -q -O $procedure >/dev/null || die_error "Could not download procedure $2" 
 	else
 		log "Loading procedure $1/procedures/$2 ..."
-		procedure=$LIB_USER/"$1"/procedures/"$2"
+		local procedure=$LIB_USER/"$1"/procedures/"$2"
 		[ "$1" = core ] && procedure=$LIB_CORE/procedures/"$2"
 	fi
 	[ -f "$procedure" ] && source "$procedure" || die_error "Something went wrong while sourcing procedure $procedure"
@@ -77,7 +79,7 @@ load_lib ()
 	[ -z "$1" ] && die_error "load_library needs a module als \$1 and library as \$2"
 	[ -z "$2" ] && die_error "load_library needs a library as \$2"
 	log "Loading library $1/libs/$2 ..."
-	lib=$LIB_USER/"$1"/libs/"$2"
+	local lib=$LIB_USER/"$1"/libs/"$2"
 	[ "$1" = core ] && lib=$LIB_CORE/libs/"$2"
 	source $lib || die_error "Something went wrong while sourcing library $lib"
 }
@@ -90,8 +92,8 @@ execute ()
 {
 	[ -z "$1" -o -z "$2" ] && debug 'MAIN' "execute $@" && die_error "Use the execute function like this: execute <type> <name> with type=phase/worker"
 	[ "$1" != phase -a "$1" != worker ] && debug 'MAIN' "execute $@" && die_error "execute's first argument must be a valid type (phase/worker)"
-	PWD_BACKUP=`pwd`
-	object=$1_$2
+	local PWD_BACKUP=`pwd`
+	local object=$1_$2
 
 	if [ "$1" = worker ]
 	then
@@ -101,7 +103,7 @@ execute ()
 			shift 2
 			$object "$@"
 			local ret=$?
-			exit_var=exit_$object
+			local exit_var=exit_$object
 			read $exit_var <<< $ret # maintain exit status of each worker
 		else
 			die_error "$object is not defined!"
@@ -109,7 +111,7 @@ execute ()
 	elif [ "$1" = phase ]
 	then
 		log "******* Executing phase $2"
-		exit_var=exit_$object
+		local exit_var=exit_$object
 		read $exit_var <<< 0
 		# TODO: for some reason the hack below does not work (tested in virtualbox), even though it really should.  Someday I must get indirect array variables working and clean this up...
 		# debug 'MAIN' "\$1: $1, \$2: $2, \$object: $object, \$exit_$object: $exit_object"
@@ -168,34 +170,40 @@ depend_procedure ()
 
 start_process ()
 {
-	execute phase preparation
-	execute phase basics
-	execute phase system
-	execute phase finish
+	ret=0
+	execute phase preparation || ret=$?
+	execute phase basics || ret=$?
+	execute phase system || ret=$?
+	execute phase finish || ret=$?
+	return $ret
 }
 
 
-show_report () #TODO: abstract UI method (cli/dia)
+show_report ()
 {
-	echo "Execution Report:"
-	echo "-----------------"
+	data="Execution Report:"
+	data="$data\n-----------------"
 	for phase in preparation basics system finish
 	do
 		object=phase_$phase
 		exit_var=exit_$object
 		local ret=${!exit_var}
-		echo -n "Phase $phase: "
-		[ "$ret" = "0" ] && echo "Success" || echo "Failed"
+		[ "$ret" = "0" ] && data="$data\nPhase $phase: Success"
+		[ "$ret" = "0" ] || data="$data\nPhase $phase: Failed"
 		eval phase_array=$(declare | grep -e "^${object}=" | cut -d"=" -f 2-)
 		for worker_str in "${phase_array[@]}"
 		do
 			worker=${worker_str%% *}
 			exit_var=exit_worker_$worker
 			ret=${!exit_var}
-			echo -n " > Worker $worker: "
-			[ "$ret" = "0" ] && echo "Success" || echo "Failed"
+			case "$ret" in
+				"")  data="$data\n > Worker $worker: Skipped";;
+				"0") data="$data\n > Worker $worker: Sucess";;
+				*)   data="$data\n > Worker $worker: Failed";;
+			esac
 		done
 	done
+	inform "$data"
 }
 
 
@@ -206,11 +214,11 @@ start_installer ()
 
 
 # use this function to stop the installation procedure.
-# $1 exit code (optional)
+# $1 exit code (optional, defaults to 0)
 stop_installer ()
 {
 	log "-------------- STOPPING INSTALLATION ----------"
 	cleanup_runtime
 	[ "$var_UI_TYPE" = dia ] && clear
-	exit $1
+	exit ${1:-0}
 }
